@@ -1,20 +1,29 @@
+from collections import namedtuple
+from pathlib import Path
 from math import log2, sqrt
 from functools import partial
 from random import randrange
 
 import torch
 import torch.nn.functional as F
-from torch import nn, einsum
+from torch.optim import Adam
+from torch import nn, einsum, Tensor
 from torch.autograd import grad as torch_grad
+from torch.utils.data import DataLoader
 
 from beartype import beartype
-from beartype.typing import List, Optional, Tuple
+from beartype.typing import List, Optional, Tuple, Dict, Union, Iterable
 
 from einops import rearrange, pack, unpack, repeat, reduce
 from einops.layers.torch import Rearrange, Reduce
 
+from ema_pytorch import EMA
+
+from gigagan_pytorch.version import __version__
 from gigagan_pytorch.open_clip import OpenClipAdapter
 
+from tqdm import tqdm
+
 # helpers
 
 def exists(val):
@@ -37,6 +46,14 @@ def safe_unshift(arr):
 def divisible_by(numer, denom):
     return (numer % denom) == 0
 
+def is_unique(arr):
+    return len(set(arr)) == len(arr)
+
+def cycle(dl):
+    while True:
+        for data in dl:
+            yield data
+
 # activation functions
 
 def leaky_relu(neg_slope = 0.1):
@@ -52,28 +69,34 @@ def log(t, eps = 1e-20):
 
 def gradient_penalty(
     images,
-    output,
+    outputs,
+    grad_output_weights = None,
     weight = 10
 ):
-    images.shape[0]
+    if not isinstance(outputs, (list, tuple)):
+        outputs = [outputs]
+
+    if not exists(grad_output_weights):
+        grad_output_weights = (1,) * len(outputs)
+
     gradients, *_ = torch_grad(
-        outputs = output,
+        outputs = outputs,
         inputs = images,
-        grad_outputs = torch.ones_like(output),
+        grad_outputs = [(torch.ones_like(output) * weight) for output, weight in zip(outputs, grad_output_weights)],
         create_graph = True,
         retain_graph = True,
         only_inputs = True
     )
 
     gradients = rearrange(gradients, 'b ... -> b (...)')
-    return weight * ((gradients.norm(2, dim=1) - 1) ** 2).mean()
+    return weight * ((gradients.norm(2, dim = 1) - 1) ** 2).mean()
 
 # hinge gan losses
 
-def gen_hinge_loss(fake):
+def generator_hinge_loss(fake):
     return fake.mean()
 
-def hinge_loss(real, fake):
+def discriminator_hinge_loss(real, fake):
     return (F.relu(1 + real) + F.relu(1 - fake)).mean()
 
 # auxiliary losses
@@ -84,9 +107,9 @@ def aux_matching_loss(real, fake):
 @beartype
 def aux_clip_loss(
     clip: OpenClipAdapter,
-    images: torch.Tensor,
+    images: Tensor,
     texts: Optional[List[str]] = None,
-    text_embeds: Optional[torch.Tensor] = None
+    text_embeds: Optional[Tensor] = None
 ):
     assert exists(texts) ^ exists(text_embeds)
 
@@ -202,8 +225,8 @@ class AdaptiveConv2DMod(nn.Module):
     def forward(
         self,
         fmap,
-        mod: Optional[torch.Tensor] = None,
-        kernel_mod: Optional[torch.Tensor] = None
+        mod: Optional[Tensor] = None,
+        kernel_mod: Optional[Tensor] = None
     ):
         """
         notation
@@ -317,7 +340,11 @@ class SelfAttention(nn.Module):
         if self.dot_product:
             sim = einsum('b i d, b j d -> b i j', q, k)
         else:
-            sim = -(torch.cdist(q, k, p = 2) ** 2)
+            # using pytorch cdist leads to nans in lightweight gan training framework, at least
+            q_squared = (q * q).sum(dim = -1)
+            k_squared = (k * k).sum(dim = -1)
+            l2dist_squared = rearrange(q_squared, 'b i -> b i 1') + rearrange(k_squared, 'b j -> b 1 j') - 2 * einsum('b i d, b j d -> b i j', q, k) # hope i'm mathing right
+            sim = -l2dist_squared
 
         # scale
 
@@ -401,19 +428,15 @@ class TextAttention(nn.Module):
         self,
         dim,
         dim_head = 64,
-        heads = 8,
-        mask_self_value = -1e2
+        heads = 8
     ):
         super().__init__()
         self.heads = heads
         self.scale = dim_head ** -0.5
         dim_inner = dim_head * heads
 
-        self.mask_self_value = mask_self_value
-
         self.norm = RMSNorm(dim)
-        self.to_qk = nn.Linear(dim, dim_inner, bias = False)
-        self.to_v = nn.Linear(dim, dim_inner, bias = False)
+        self.to_qkv = nn.Linear(dim, dim_inner * 3, bias = False)
 
         self.null_kv = nn.Parameter(torch.randn(2, heads, dim_head))
 
@@ -431,16 +454,14 @@ class TextAttention(nn.Module):
         i - source seq (attend from)
         j - target seq (attend to)
         """
-        batch, device = encodings.shape[0], encodings.device
+        batch = encodings.shape[0]
 
         encodings = self.norm(encodings)
 
         h = self.heads
 
-        qk, v = self.to_qk(encodings), self.to_v(encodings)
-        qk, v = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h = self.heads), (qk, v))
-
-        q, k = qk, qk
+        q, k, v = self.to_qkv(encodings).chunk(3, dim = -1)
+        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h = self.heads), (q, k, v))
 
         # add a null key / value, so network can choose to pay attention to nothing
 
@@ -449,17 +470,7 @@ class TextAttention(nn.Module):
         k = torch.cat((nk, k), dim = -2)
         v = torch.cat((nv, v), dim = -2)
 
-        # l2 distance
-
-        sim = -torch.cdist(q, k, p = 2) * self.scale
-
-        # following what was done in reformer for shared query / key space
-        # omit attention to self
-
-        self_mask = torch.eye(sim.shape[-2], device = device, dtype = torch.bool)
-        self_mask = F.pad(self_mask, (1, 0), value = False)
-
-        sim = sim.masked_fill(self_mask, self.mask_self_value)
+        sim = einsum('b i d, b j d -> b i j', q, k) * self.scale
 
         # key padding mask
 
@@ -593,11 +604,15 @@ class TextEncoder(nn.Module):
     @beartype
     def forward(
         self,
-        texts: List[str]
+        texts: Optional[List[str]] = None,
+        text_encodings: Optional[Tensor] = None
     ):
-        with torch.no_grad():
-            self.clip.eval()
-            _, text_encodings = self.clip.embed_texts(texts)
+        assert exists(texts) ^ exists(text_encodings)
+
+        if not exists(text_encodings):
+            with torch.no_grad():
+                self.clip.eval()
+                _, text_encodings = self.clip.embed_texts(texts)
 
         mask = (text_encodings != 0.).any(dim = -1)
 
@@ -618,13 +633,31 @@ class TextEncoder(nn.Module):
 
 # style mapping network
 
+class EqualLinear(nn.Module):
+    def __init__(
+        self,
+        dim,
+        dim_out,
+        lr_mul = 1,
+        bias = True
+    ):
+        super().__init__()
+        self.weight = nn.Parameter(torch.randn(dim_out, dim))
+        if bias:
+            self.bias = nn.Parameter(torch.zeros(dim_out))
+
+        self.lr_mul = lr_mul
+
+    def forward(self, input):
+        return F.linear(input, self.weight * self.lr_mul, bias=self.bias * self.lr_mul)
+
 class StyleNetwork(nn.Module):
     def __init__(
         self,
         dim,
         depth,
-        dim_text_latent = 0,
-        frac_gradient = 0.1  # in the stylegan2 paper, they control the learning rate by multiplying the parameters by a constant, but we can use another trick here from attention literature
+        lr_mul = 0.1,
+        dim_text_latent = 0
     ):
         super().__init__()
         self.dim = dim
@@ -633,44 +666,43 @@ class StyleNetwork(nn.Module):
         layers = []
         for i in range(depth):
             is_first = i == 0
+            dim_in = (dim + dim_text_latent) if is_first else dim
 
-            dim_in = dim + (dim_text_latent if is_first else 0)
-            layers.extend([nn.Linear(dim_in, dim), leaky_relu()])
+            layers.extend([EqualLinear(dim_in, dim, lr_mul), leaky_relu()])
 
         self.net = nn.Sequential(*layers)
-        self.frac_gradient = frac_gradient
-        self.dim_text_latent = dim_text_latent
 
     def forward(
         self,
         x,
         text_latent = None
     ):
-        grad_frac = self.frac_gradient
+        x = F.normalize(x, dim = 1)
 
-        if self.dim_text_latent:
+        if self.dim_text_latent > 0:
             assert exists(text_latent)
             x = torch.cat((x, text_latent), dim = -1)
 
-        x = F.normalize(x, dim = 1)
-        out = self.net(x)
-
-        return out * grad_frac + (1 - grad_frac) * out.detach()
+        return self.net(x)
 
 # generator
 
-class Generator(nn.Module):
+class BaseGenerator(nn.Module):
+    pass
+
+class Generator(BaseGenerator):
     @beartype
     def __init__(
         self,
         *,
         dim,
         image_size,
-        dim_max = 8192,
+        dim_max = 2048,
         capacity = 16,
         channels = 3,
-        style_network: Optional[StyleNetwork] = None,
-        text_encoder: Optional[TextEncoder] = None,
+        style_network: Optional[Union[StyleNetwork, Dict]] = None,
+        style_network_dim = None,
+        text_encoder: Optional[Union[TextEncoder, Dict]] = None,
         dim_latent = 512,
         self_attn_resolutions: Tuple[int, ...] = (32, 16),
         self_attn_dim_head = 64,
@@ -682,7 +714,6 @@ class Generator(nn.Module):
         cross_attn_heads = 8,
         cross_ff_mult = 4,
         num_conv_kernels = 2,  # the number of adaptive conv kernels
-        use_glu = False,
         num_skip_layers_excite = 0,
         unconditional = False
     ):
@@ -690,7 +721,21 @@ class Generator(nn.Module):
         self.dim = dim
         self.channels = channels
 
+        if isinstance(style_network, dict):
+            style_network = StyleNetwork(**style_network)
+
         self.style_network = style_network
+
+        assert exists(style_network) ^ exists(style_network_dim), 'style_network_dim must be given to the generator if StyleNetwork not passed in as style_network'
+
+        if not exists(style_network_dim):
+            style_network_dim = style_network.dim
+
+        self.style_network_dim = style_network_dim
+
+        if isinstance(text_encoder, dict):
+            text_encoder = TextEncoder(**text_encoder)
+
         self.text_encoder = text_encoder
 
         self.unconditional = unconditional
@@ -760,14 +805,11 @@ class Generator(nn.Module):
                 dim_skip_in, _ = dim_pairs[ind + num_skip_layers_excite]
                 skip_squeeze_excite = SqueezeExcite(dim_in, dim_skip_in)
 
-            mult = 2 if use_glu else 1
-            act_fn = partial(nn.GLU, dim = 1) if use_glu else leaky_relu
-
             resnet_block = nn.ModuleList([
-                adaptive_conv(dim_in, dim_out * mult),
-                act_fn(),
-                adaptive_conv(dim_out, dim_out * mult),
-                act_fn()
+                adaptive_conv(dim_in, dim_out),
+                leaky_relu(),
+                adaptive_conv(dim_out, dim_out),
+                leaky_relu()
             ])
 
             to_rgb = adaptive_conv(dim_out, channels)
@@ -805,9 +847,15 @@ class Generator(nn.Module):
 
         # determine the projection of the style embedding to convolutional modulation weights (+ adaptive kernel selection weights) for all layers
 
-        self.style_to_conv_modulations = nn.Linear(style_network.dim, sum(style_embed_split_dims))
+        self.style_to_conv_modulations = nn.Linear(style_network_dim, sum(style_embed_split_dims))
         self.style_embed_split_dims = style_embed_split_dims
 
+        self.apply(self.init_)
+
+    def init_(self, m):
+        if type(m) in {nn.Conv2d, nn.Linear}:
+            nn.init.kaiming_normal_(m.weight, a = 0, mode = 'fan_in', nonlinearity = 'leaky_relu')
+
     @property
     def device(self):
         return next(self.parameters()).device
@@ -833,7 +881,7 @@ class Generator(nn.Module):
                 assert exists(self.text_encoder)
                 global_text_tokens, fine_text_tokens, text_mask = self.text_encoder(texts)
             else:
-                assert all([*map(exists, (global_text_tokens, fine_text_tokens, text_mask))])
+                assert all([*map(exists, (global_text_tokens, fine_text_tokens, text_mask))]), 'raw text or text embeddings were not passed in for conditional training'
         else:
             assert not any([*map(exists, (texts, global_text_tokens, fine_text_tokens))])
 
@@ -841,7 +889,10 @@ class Generator(nn.Module):
 
         if not exists(styles):
             assert exists(self.style_network)
-            noise = default(noise, torch.randn((batch_size, self.dim), device = self.device))
+
+            if not exists(noise):
+                noise = torch.randn((batch_size, self.style_network_dim), device = self.device)
+
             styles = self.style_network(noise, global_text_tokens)
 
         # project styles to conv modulations
@@ -1104,24 +1155,24 @@ class Discriminator(nn.Module):
         dim,
         image_size,
         capacity = 16,
-        dim_max = 8192,
+        dim_max = 2048,
         channels = 3,
-        attn_resolutions: Tuple[int] = (32, 16),
+        attn_resolutions: Tuple[int, ...] = (32, 16),
         attn_dim_head = 64,
         attn_heads = 8,
         self_attn_dot_product = False,
         ff_mult = 4,
-        text_encoder: Optional[TextEncoder] = None,
+        text_encoder: Optional[Union[TextEncoder, Dict]] = None,
         text_dim = None,
-        multiscale_input_resolutions: Tuple[int] = (64, 32, 16, 8),
-        multiscale_output_resolutions: Tuple[int] = (32, 16, 8, 4),
-        aux_recon_resolutions: Tuple[int] = (8,),
-        aux_recon_patches: Tuple[int] = (2,),
+        multiscale_input_resolutions: Tuple[int, ...] = (64, 32, 16, 8),
+        multiscale_output_resolutions: Tuple[int, ...] = (32, 16, 8, 4),
+        aux_recon_resolutions: Tuple[int, ...] = (8,),
+        aux_recon_patches: Tuple[int, ...] = (2,),
         resize_mode = 'bilinear',
         num_conv_kernels = 2,
-        use_glu = False,
         num_skip_layers_excite = 0,
-        unconditional = False
+        unconditional = False,
+        scale_invariant_training = True
     ):
         super().__init__()
         self.unconditional = unconditional
@@ -1130,15 +1181,21 @@ class Discriminator(nn.Module):
         assert is_power_of_two(image_size)
         assert all([*map(is_power_of_two, attn_resolutions)])
 
+        assert is_unique(multiscale_input_resolutions)
         assert all([*map(is_power_of_two, multiscale_input_resolutions)])
         assert all([*map(lambda t: t >= 4, multiscale_input_resolutions)])
+        assert all([*map(lambda t: t < image_size, multiscale_input_resolutions)])
+
         self.multiscale_input_resolutions = multiscale_input_resolutions
 
+        assert is_unique(multiscale_output_resolutions)
         assert all([*map(is_power_of_two, multiscale_output_resolutions)])
         assert all([*map(lambda t: t >= 4, multiscale_output_resolutions)])
+        assert all([*map(lambda t: t < image_size, multiscale_output_resolutions)])
 
-        assert max(multiscale_input_resolutions) > max(multiscale_output_resolutions)
-        assert min(multiscale_input_resolutions) > min(multiscale_output_resolutions)
+        if len(multiscale_input_resolutions) > 0 and len(multiscale_output_resolutions) > 0:
+            assert max(multiscale_input_resolutions) > max(multiscale_output_resolutions)
+            assert min(multiscale_input_resolutions) > min(multiscale_output_resolutions)
 
         self.multiscale_output_resolutions = multiscale_output_resolutions
 
@@ -1150,6 +1207,7 @@ class Discriminator(nn.Module):
 
         num_layers = int(log2(image_size) - 1)
         self.num_layers = num_layers
+        self.image_size = image_size
 
         resolutions = image_size / ((2 ** torch.arange(num_layers)))
         resolutions = resolutions.long().tolist()
@@ -1192,7 +1250,7 @@ class Discriminator(nn.Module):
             # multi-scale rgb input to feature dimension
 
             from_rgb = None
-            if has_multiscale_input:
+            if has_multiscale_input and scale_invariant_training:
                 from_rgb = nn.Conv2d(channels, dim_in, 7, padding = 3)
 
             # dim in + channels for the main features
@@ -1205,19 +1263,17 @@ class Discriminator(nn.Module):
 
             # main resnet block
 
-            mult = 2 if use_glu else 1
-            act_fn = partial(nn.GLU, dim = 1) if use_glu else leaky_relu
-
             resnet_block = nn.Sequential(
-                conv2d_3x3(dim_in, dim_out * mult),
-                act_fn(),
-                conv2d_3x3(dim_out, dim_out * mult),
-                act_fn()
+                conv2d_3x3(dim_in, dim_out),
+                leaky_relu(),
+                conv2d_3x3(dim_out, dim_out),
+                leaky_relu()
             )
 
             # multi-scale output
 
             multiscale_output_predictor = None
+
             if has_multiscale_output:
                 multiscale_output_predictor = Predictor(dim_out, num_conv_kernels = num_conv_kernels, unconditional = unconditional)
                 predictor_dims.extend([dim_out, dim_kernel_attn])
@@ -1257,13 +1313,22 @@ class Discriminator(nn.Module):
         assert unconditional or (exists(text_dim) ^ exists(text_encoder))
 
         if not unconditional:
-            self.text_encoder = text_encoder
+            if isinstance(text_encoder, dict):
+                text_encoder = TextEncoder(**text_encoder)
 
             self.text_dim = default(text_dim, text_encoder.dim)
 
             self.predictor_dims = predictor_dims
             self.text_to_conv_conditioning = nn.Linear(self.text_dim, sum(predictor_dims)) if exists(self.text_dim) else None
 
+        self.text_encoder = text_encoder
+
+        self.apply(self.init_)
+
+    def init_(self, m):
+        if type(m) in {nn.Conv2d, nn.Linear}:
+            nn.init.kaiming_normal_(m.weight, a = 0, mode = 'fan_in', nonlinearity = 'leaky_relu')
+
     def resize_image_to(self, images, resolution):
         return F.interpolate(images, resolution, mode = self.resize_mode)
 
@@ -1271,10 +1336,12 @@ class Discriminator(nn.Module):
     def forward(
         self,
         images,
-        rgbs: Optional[List[torch.Tensor]] = None,  # multi-resolution inputs (rgbs) from the generator
+        rgbs: Optional[List[Tensor]] = None,  # multi-resolution inputs (rgbs) from the generator
         texts: Optional[List[str]] = None,
         text_embeds = None,
-        real_images = None                          # if this were passed in, the network will automatically append the real to the presumably generated images passed in as the first argument, and generate all intermediate resolutions through resizing and concat appropriately
+        real_images = None,                   # if this were passed in, the network will automatically append the real to the presumably generated images passed in as the first argument, and generate all intermediate resolutions through resizing and concat appropriately
+        return_all_aux_loss = False           # this would return auxiliary reconstruction loss for both fake and real
+
     ):
         if not self.unconditional:
             assert exists(texts) ^ exists(text_embeds)
@@ -1283,21 +1350,29 @@ class Discriminator(nn.Module):
                 assert exists(self.text_encoder)
                 text_embeds, *_ = self.text_encoder(texts)
 
+            assert exists(text_embeds), 'raw text or text embeddings were not passed into discriminator for conditional training'
+
             conv_mods = self.text_to_conv_conditioning(text_embeds).split(self.predictor_dims, dim = -1)
             conv_mods = iter(conv_mods)
+
         else:
             assert not any([*map(exists, (texts, text_embeds))])
 
-        batch = images.shape[0]
-
         x = images
 
+        assert x.shape[-2:] == (self.image_size, self.image_size)
+
         # if real images are passed in, assume `images` are generated, and take care of all the multi-resolution input. this can also be done externally, in which case `real_images` will not be populated
 
         has_real_images = exists(real_images)
-        x = torch.cat((x, real_images), dim = 0)
 
-        aux_recon_target = x
+        if has_real_images:
+            split_batch_size = (x.shape[0], real_images.shape[0])
+            x = torch.cat((x, real_images), dim = 0)
+
+        batch = x.shape[0]
+
+        aux_recon_target = real_images if has_real_images else x
 
         assert not (has_real_images and not exists(rgbs)) 
 
@@ -1372,27 +1447,445 @@ class Discriminator(nn.Module):
             x = x + residual
             x = x * self.residual_scale
 
-            if exists(recon_decoder):
-                aux_recon_target = repeat(aux_recon_target, 'b ... -> (s b) ...', s = x.shape[0] // aux_recon_target.shape[0])
-                aux_recon_loss = recon_decoder(x, aux_recon_target)
+            if exists(recon_decoder) and (return_all_aux_loss or has_real_images):
+
+                if return_all_aux_loss:
+                    recon_output = x
+
+                elif has_real_images:
+                    recon_output = rearrange(x, '(s b) ... -> s b ...', b = batch)
+                    _, recon_output = recon_output.split(split_batch_size, dim = 1)
+                    recon_output = rearrange(recon_output, 's b ... -> (s b) ...')
+
+                aux_recon_target = repeat(aux_recon_target, 'b ... -> (s b) ...', s = recon_output.shape[0] // aux_recon_target.shape[0])
+                aux_recon_loss = recon_decoder(recon_output, aux_recon_target)
                 aux_recon_losses.append(aux_recon_loss)
 
         logits = self.to_logits(x)   
         logits = rearrange(logits, '(s b) ... -> s b ...', b = batch)
 
-        return logits, multiscale_outputs, aux_recon_losses
+        if not has_real_images:
+            return logits, multiscale_outputs, aux_recon_losses
+
+        # if real images are present, break up the outputs into (fake, real) tuples
+
+        split_logits = logits.split(split_batch_size, dim = 1)
+
+        split_multiscale_outputs = []
+
+        for multiscale_output in multiscale_outputs:
+            multiscale_output = rearrange(multiscale_output, '(s b) ... -> s b ...', b = batch)
+            multiscale_output = multiscale_output.split(split_batch_size, dim = 1)
+            split_multiscale_outputs.append(multiscale_output)
+
+        return split_logits, split_multiscale_outputs, aux_recon_losses
 
 # gan
 
+TrainDiscrLosses = namedtuple('TrainDiscrLosses', [
+    'divergence',
+    'multiscale_divergence',
+    'gradient_penalty',
+    'aux_reconstruction'
+])
+
+TrainGenLosses = namedtuple('TrainGenLosses', [
+    'divergence',
+    'multiscale_divergence'
+])
+
 class GigaGAN(nn.Module):
     @beartype
     def __init__(
         self,
         *,
-        generator: Generator,
-        discriminator: Discriminator
+        generator: Union[BaseGenerator, Dict],
+        discriminator: Union[Discriminator, Dict],
+        text_encoder: Optional[Union[TextEncoder, Dict]] = None,
+        learning_rate = 1e-4,
+        betas = (0.9, 0.99),
+        discr_aux_recon_loss_weight = 0.25,
+        multiscale_divergence_loss_weight = 1.,
+        apply_gradient_penalty_every = 16,
+        train_upsampler = False,
+        upsampler_replace_rgb_with_input_lowres_image = False,
+        log_steps_every = 20,
+        create_ema_generator_at_init = True
     ):
         super().__init__()
 
-    def forward(self, x):
-        return x
+        self.train_upsampler = train_upsampler
+
+        self.upsampler_replace_rgb_with_input_lowres_image= upsampler_replace_rgb_with_input_lowres_image
+        self.apply_gradient_penalty_every = apply_gradient_penalty_every
+
+        if train_upsampler:
+            from gigagan_pytorch.unet_upsampler import UnetUpsampler
+            generator_klass = UnetUpsampler
+        else:
+            generator_klass = Generator
+
+        if isinstance(generator, dict):
+            generator = generator_klass(**generator)
+
+        if isinstance(discriminator, dict):
+            discriminator = Discriminator(**discriminator)
+
+        assert isinstance(generator, generator_klass)
+
+        self.G = generator
+        self.D = discriminator
+
+        # text encoder
+
+        if exists(text_encoder):
+            if isinstance(text_encoder, dict):
+                text_encoder = TextEncoder(**text_encoder)
+
+        self.text_encoder = text_encoder
+
+        assert generator.unconditional == discriminator.unconditional
+
+        self.unconditional = generator.unconditional
+
+        # optimizers
+
+        self.G_opt = Adam(self.G.parameters(), lr = learning_rate, betas = betas)
+        self.D_opt = Adam(self.D.parameters(), lr = learning_rate, betas = betas)
+
+        # loss related
+
+        self.discr_aux_recon_loss_weight = discr_aux_recon_loss_weight
+        self.multiscale_divergence_loss_weight = multiscale_divergence_loss_weight
+
+        # ema
+
+        self.has_ema_generator = False
+
+        if create_ema_generator_at_init:
+            self.create_ema_generator()
+
+        # steps
+
+        self.log_steps_every = log_steps_every
+
+        self.register_buffer('steps', torch.ones(1, dtype = torch.long))
+
+    def save(self, path, overwrite = True):
+        path = Path(path)
+        path.parents[0].mkdir(parents = True, exist_ok =True)
+
+        assert overwrite or not path.exists()
+
+        pkg = dict(
+            G = self.G.state_dict(),
+            D = self.D.state_dict(),
+            G_opt = self.G_opt.state_dict(),
+            D_opt = self.D_opt.state_dict(),
+            steps = self.steps.item(),
+            version = __version__
+        )
+
+        if self.has_ema_generator:
+            pkg['G_ema'] = self.G_ema.state_dict()
+
+        torch.save(pkg, str(path))
+
+    def load(self, path, strict = False):
+        path = Path(path)
+        assert path.exists()
+
+        pkg = torch.load(str(path))
+
+        if 'version' in pkg and pkg['version'] != __version__:
+            print(f"trying to load from version {pkg['version']}")
+
+        self.G.load_state_dict(pkg['G'], strict = strict)
+        self.D.load_state_dict(pkg['D'], strict = strict)
+
+        if self.has_ema_generator:
+            self.G_ema.load_state_dict(pkg['G_ema'])
+
+        if 'steps' in pkg:
+            self.steps.copy_(torch.tensor([pkg['steps']]))
+
+        if 'G_opt'not in pkg or 'D_opt' not in pkg:
+            return
+
+        try:
+            self.G_opt.load_state_dict(pkg['G_opt'])
+            self.D_opt.load_state_dict(pkg['D_opt'])
+        except Exception as e:
+            self.print(f'unable to load optimizers {e.msg}- optimizer states will be reset')
+            pass
+
+    @property
+    def device(self):
+        return self.steps.device
+
+    def create_ema_generator(
+        self,
+        update_every = 10,
+        update_after_step = 100,
+        decay = 0.995
+    ):
+        assert not self.has_ema_generator, 'EMA generator has already been created'
+
+        self.has_ema_generator = True
+        self.G_ema = EMA(self.G, update_every = update_every, update_after_step = update_after_step, beta = decay)
+
+    def print(self, msg):
+        print(msg)
+
+    @beartype
+    def train_discriminator_step(
+        self,
+        dl_iter: Iterable,
+        grad_accum_every = 1,
+        apply_gradient_penalty = False
+    ):
+        total_divergence = 0.
+        total_gp_loss = 0.
+        total_aux_loss = 0.
+        total_multiscale_divergence = 0.
+
+        self.D_opt.zero_grad()
+
+        for _ in range(grad_accum_every):
+
+            if self.unconditional:
+                real_images = next(dl_iter)
+            else:
+                result = next(dl_iter)
+                assert isinstance(result, tuple), 'dataset should return a tuple of two items for text conditioned training, (images: Tensor, texts: List[str])'
+                real_images, texts = result
+
+            # requires grad for real images, for gradient penalty
+
+            real_images = real_images.to(self.device)
+            real_images.requires_grad_()
+
+            batch_size = real_images.shape[0]
+
+            # for discriminator training, fit upsampler and image synthesis logic under same function
+
+            if self.train_upsampler:
+                size = self.G.input_image_size
+                lowres_real_images = F.interpolate(real_images, (size, size))
+
+                G_kwargs = dict(
+                    lowres_image = lowres_real_images,
+                    replace_rgb_with_input_lowres_image = self.upsampler_replace_rgb_with_input_lowres_image
+                )
+            else:
+                G_kwargs = dict(batch_size = batch_size)
+
+            # add texts for conditioning if needed
+
+            maybe_text_kwargs = dict()
+
+            if not self.unconditional:
+                maybe_text_kwargs['texts'] = texts
+
+            # generator
+
+            with torch.no_grad():
+                images, rgbs = self.G(
+                    **G_kwargs,
+                    **maybe_text_kwargs,
+                    return_all_rgbs = True
+                )
+
+            # detach output of generator, as training discriminator only
+
+            images.detach_()
+
+            for rgb in rgbs:
+                rgb.detach_()
+
+            # main divergence loss
+
+            (fake_logits, real_logits), multiscale_logits, aux_recon_losses = self.D(
+                images,
+                rgbs,
+                real_images = real_images,
+                **maybe_text_kwargs,
+            )
+
+            divergence = discriminator_hinge_loss(real_logits, fake_logits)
+            total_divergence = total_divergence + (divergence / grad_accum_every)
+
+            # handle multi-scale divergence
+
+            multiscale_divergence = 0.
+
+            multiscale_real_logits = []
+
+            if self.multiscale_divergence_loss_weight > 0.:
+
+                for multiscale_fake, multiscale_real in multiscale_logits:
+                    multiscale_loss = discriminator_hinge_loss(multiscale_fake, multiscale_real)
+                    multiscale_divergence = multiscale_divergence + multiscale_loss
+                    multiscale_real_logits.append(multiscale_real)
+
+                total_multiscale_divergence += multiscale_divergence / grad_accum_every
+
+            # figure out gradient penalty if needed
+
+            gp_loss = 0.
+
+            if apply_gradient_penalty:
+                gp_loss = gradient_penalty(
+                    real_images,
+                    outputs = [real_logits, *multiscale_real_logits],
+                    grad_output_weights = [1., *(self.multiscale_divergence_loss_weight,) * len(multiscale_real_logits)]
+                )
+
+                total_gp_loss += gp_loss / grad_accum_every
+
+            # sum up losses
+
+            total_loss = divergence + gp_loss
+
+            if self.multiscale_divergence_loss_weight > 0.:
+                total_loss = total_loss + multiscale_divergence * self.multiscale_divergence_loss_weight
+
+            if self.discr_aux_recon_loss_weight > 0.:
+                aux_loss = sum(aux_recon_losses)
+
+                total_aux_loss += aux_loss / grad_accum_every
+
+                total_loss = total_loss + aux_loss * self.discr_aux_recon_loss_weight
+
+            # backwards
+
+            (total_loss / grad_accum_every).backward()
+
+        self.D_opt.step()
+
+        return TrainDiscrLosses(total_divergence, total_multiscale_divergence, total_gp_loss, total_aux_loss)
+
+    def train_generator_step(
+        self,
+        batch_size = None,
+        dl_iter: Optional[Iterable] = None,
+        grad_accum_every = 1
+    ):
+        total_divergence = 0.
+        total_multiscale_divergence = 0.
+
+        self.D_opt.zero_grad()
+        self.G_opt.zero_grad()
+
+        for _ in range(grad_accum_every):
+
+            # what to pass into the generator
+            # depends on whether training upsampler or not
+
+            maybe_text_kwargs = dict()
+
+            if self.train_upsampler or not self.unconditional:
+                assert exists(dl_iter)
+
+                if self.unconditional:
+                    real_images = next(dl_iter)
+                else:
+                    result = next(dl_iter)
+                    assert isinstance(result, tuple), 'dataset should return a tuple of two items for text conditioned training, (images: Tensor, texts: List[str])'
+                    real_images, texts = result
+
+                    maybe_text_kwargs['texts'] = texts
+
+                real_images = real_images.to(self.device)
+
+            # if training upsample generator, need to downsample real images
+
+            if self.train_upsampler:
+                size = self.G.input_image_size
+                lowres_real_images = F.interpolate(real_images, (size, size))
+
+                G_kwargs = dict(
+                    lowres_image = lowres_real_images,
+                    replace_rgb_with_input_lowres_image = self.upsampler_replace_rgb_with_input_lowres_image
+                )
+            else:
+                assert exists(batch_size)
+
+                G_kwargs = dict(batch_size = batch_size)
+
+            # generator
+
+            image, rgbs = self.G(
+                **G_kwargs,
+                **maybe_text_kwargs,
+                return_all_rgbs = True
+            )
+
+            # discriminator
+
+            logits, multiscale_logits, _ = self.D(
+                image,
+                rgbs,
+                **maybe_text_kwargs
+            )
+
+            # hinge loss
+
+            divergence = generator_hinge_loss(logits)
+
+            total_divergence += divergence
+
+            total_loss = divergence
+
+            if self.multiscale_divergence_loss_weight > 0.:
+                multiscale_divergence = 0.
+
+                for multiscale_logit in multiscale_logits:
+                    multiscale_divergence = multiscale_divergence + generator_hinge_loss(multiscale_logit)
+
+                total_multiscale_divergence += multiscale_divergence
+
+                total_loss = total_loss + multiscale_divergence * self.multiscale_divergence_loss_weight
+
+            (total_loss / grad_accum_every).backward()
+
+        self.G_opt.step()
+
+        # update exponentially moving averaged generator
+
+        if self.has_ema_generator:
+            self.G_ema.update()
+
+        return TrainGenLosses(total_divergence, total_multiscale_divergence)
+
+    @beartype
+    def forward(
+        self,
+        *,
+        steps,
+        dataloader: DataLoader,
+        grad_accum_every = 1
+    ):
+        batch_size = dataloader.batch_size
+        dl_iter = cycle(dataloader)
+
+        last_gp_loss = 0.
+
+        for _ in tqdm(range(steps), initial = self.steps.item()):
+            steps = self.steps.item()
+            is_first_step = steps == 1
+
+            apply_gradient_penalty = self.apply_gradient_penalty_every > 0 and divisible_by(steps, self.apply_gradient_penalty_every)
+
+            d_loss, multiscale_d_loss, gp_loss, recon_loss = self.train_discriminator_step(dl_iter, grad_accum_every = grad_accum_every, apply_gradient_penalty = apply_gradient_penalty)
+            g_loss, multiscale_g_loss = self.train_generator_step(dl_iter = dl_iter, batch_size = batch_size, grad_accum_every = grad_accum_every)
+
+            if exists(gp_loss):
+                last_gp_loss = gp_loss
+
+            if is_first_step or divisible_by(steps, self.log_steps_every):
+                self.print(f' G: {g_loss:.2f} | MSG: {multiscale_g_loss:.2f} | D: {d_loss:.2f} | MSD: {multiscale_d_loss:.2f} | GP: {last_gp_loss:.2f} | SSL: {recon_loss:.2f}')
+
+            self.steps += 1
+
+        self.print(f'complete {steps} training steps')
