@@ -10,6 +10,7 @@ from einops.layers.torch import Rearrange
 
 from gigagan_pytorch.attend import Attend
 from gigagan_pytorch.gigagan_pytorch import (
+    BaseGenerator,
     StyleNetwork,
     AdaptiveConv2DMod,
     TextEncoder,
@@ -17,8 +18,7 @@ from gigagan_pytorch.gigagan_pytorch import (
 )
 
 from beartype import beartype
-from beartype.typing import Optional, List
-from collections.abc import Iterable
+from beartype.typing import Optional, List, Union, Dict, Iterable
 
 # helpers functions
 
@@ -286,7 +286,7 @@ class LinearTransformer(nn.Module):
 
 # model
 
-class UnetUpsampler(nn.Module):
+class UnetUpsampler(BaseGenerator):
 
     @beartype
     def __init__(
@@ -297,8 +297,8 @@ class UnetUpsampler(nn.Module):
         input_image_size,
         init_dim = None,
         out_dim = None,
-        text_encoder: Optional[TextEncoder] = None,
-        style_network: Optional[StyleNetwork] = None,
+        text_encoder: Optional[Union[TextEncoder, Dict]] = None,
+        style_network: Optional[Union[StyleNetwork, Dict]] = None,
         dim_mults = (1, 2, 4, 8),
         channels = 3,
         resnet_block_groups = 8,
@@ -321,7 +321,14 @@ class UnetUpsampler(nn.Module):
 
         # style network
 
+        if isinstance(text_encoder, dict):
+            text_encoder = TextEncoder(**text_encoder)
+
         self.text_encoder = text_encoder
+
+        if isinstance(style_network, dict):
+            style_network = StyleNetwork(**style_network)
+
         self.style_network = style_network
 
         # validate text conditioning and style network hparams
@@ -442,15 +449,17 @@ class UnetUpsampler(nn.Module):
 
     def forward(
         self,
-        x,
+        lowres_image,
         styles = None,
         noise = None,
         texts: Optional[List[str]] = None,
         global_text_tokens = None,
         fine_text_tokens = None,
         text_mask = None,
-        return_all_rgbs = False
+        return_all_rgbs = False,
+        replace_rgb_with_input_lowres_image = True   # discriminator should also receive the low resolution image the upsampler sees
     ):
+        x = lowres_image
         shape = x.shape
         batch_size = shape[0]
 
@@ -561,4 +570,17 @@ class UnetUpsampler(nn.Module):
 
         rgbs = list(filter(lambda t: t.shape[-1] >= shape[-1], rgbs))
 
-        return rgb, rgbs
+        if not replace_rgb_with_input_lowres_image:
+            return rgb, rgbs
+
+        # replace the rgb of the corresponding same dimension as the input low res image
+
+        output_rgbs = []
+
+        for rgb in rgbs:
+            if rgb.shape[-1] == lowres_image.shape[-1]:
+                output_rgbs.append(lowres_image)
+            else:
+                output_rgbs.append(rgb)
+
+        return rgb, output_rgbs
